import lox_space as lox
import pytest


def test_time(provider):
    tai_exp = lox.Time("TAI", 2000, 1, 1)
    tai_act = lox.Time.from_iso("2000-01-01T00:00:00.000 TAI")
    assert tai_exp == tai_act
    tai_act = tai_exp.to_tai()
    assert tai_exp == tai_act
    tai_act = tai_exp.to_tcb().to_tai()
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_tcg().to_tai()
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_tdb().to_tai()
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_tt().to_tai()
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_ut1(provider).to_tai(provider)
    assert tai_exp.isclose(tai_act)
    with pytest.raises(ValueError):
        tai_exp.to_ut1()
    tai1 = lox.Time("TAI", 2000, 1, 1, 0, 0, 0.5)
    assert tai1 > tai_exp
    assert tai1 >= tai_exp
    assert tai_exp < tai1
    assert tai_exp <= tai1
    assert tai_exp != tai1
    dt = lox.TimeDelta(0.5)
    assert tai_exp + dt == tai1
    assert tai1 - dt == tai_exp
    assert tai1 - tai_exp == dt


def test_utc(provider):
    utc_exp = lox.UTC(2000, 1, 1)
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000")
    assert utc_exp == utc_act
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000Z")
    assert utc_exp == utc_act
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000 UTC")
    assert utc_exp == utc_act
    utc_act = utc_exp.to_tai().to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_tcb().to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_tcg().to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_tdb().to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_tt().to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_ut1(provider).to_utc(provider)
    assert utc_exp == utc_act


def test_time_delta():
    delta = lox.TimeDelta(1.5)
    assert str(delta) == "1.5 seconds"
    assert delta.seconds() == 1
    assert delta.subsecond() == 0.5
    assert str(delta + delta) == "3 seconds"
    assert str(delta - delta) == "0 seconds"
    assert str(-delta) == "-1.5 seconds"
    with pytest.raises(ValueError):
        lox.TimeDelta(float("nan"))



#  Copyright (c) 2024. Helge Eichhorn and the LOX contributors
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, you can obtain one at https://mozilla.org/MPL/2.0/.

import lox_space as lox
import numpy as np
import numpy.testing as npt
import math
import pytest


@pytest.fixture
def orbit():
    utc = lox.UTC(2023, 3, 25, 21, 8, 0.0)
    time = utc.to_tdb()
    semi_major_axis = 24464.560
    eccentricity = 0.7311
    inclination = 0.122138
    longitude_of_ascending_node = 1.00681
    argument_of_periapsis = 3.10686
    true_anomaly = 0.44369564302687126
    return lox.Keplerian(time, semi_major_axis, eccentricity, inclination, longitude_of_ascending_node,
                         argument_of_periapsis, true_anomaly)


@pytest.fixture
def trajectory(orbit):
    dt = orbit.orbital_period()
    rng = [orbit.time() + dt for dt in lox.TimeDelta.range(0, math.ceil(dt))]
    s0 = orbit.to_cartesian()
    prop = lox.Vallado(s0)
    return prop.propagate(rng)


def test_from_numpy():
    utc = lox.UTC(2023, 3, 25, 21, 8, 0.0)
    time = utc.to_tdb()
    states = np.array([
        [0.0, 1e3, 1e3, 1e3, 1.0, 1.0, 1.0],
        [1.0, 2e3, 2e3, 2e3, 2.0, 2.0, 2.0],
        [2.0, 3e3, 3e3, 3e3, 3.0, 3.0, 3.0],
        [3.0, 4e3, 4e3, 4e3, 4.0, 4.0, 4.0],
    ])
    trajectory = lox.Trajectory.from_numpy(time, states)
    npt.assert_allclose(trajectory.interpolate(time + lox.TimeDelta(1.5)).position(), np.array([2.5e3, 2.5e3, 2.5e3]))
    states1 = trajectory.to_numpy()
    npt.assert_allclose(states, states1)


def test_interpolation(orbit, trajectory):
    dt = orbit.orbital_period()
    s1 = trajectory.interpolate(dt)
    k1 = s1.to_keplerian()
    assert orbit.semi_major_axis() == pytest.approx(k1.semi_major_axis(), rel=1e-8)
    assert orbit.eccentricity() == pytest.approx(k1.eccentricity(), rel=1e-8)
    assert orbit.inclination() == pytest.approx(k1.inclination(), rel=1e-8)
    assert orbit.longitude_of_ascending_node() == pytest.approx(k1.longitude_of_ascending_node(), rel=1e-8)
    assert orbit.argument_of_periapsis() == pytest.approx(k1.argument_of_periapsis(), rel=1e-8)
    assert orbit.true_anomaly() == pytest.approx(k1.true_anomaly(), rel=1e-8)


def test_events(trajectory):
    def apsis_pass(s):
        return s.position() @ s.velocity()

    events = trajectory.find_events(apsis_pass)
    assert len(events) == 2
    k1 = trajectory.interpolate(events[0].time()).to_keplerian()
    assert k1.true_anomaly() == pytest.approx(np.pi, rel=1e-8)
    k2 = trajectory.interpolate(events[1].time()).to_keplerian()
    assert k2.true_anomaly() == pytest.approx(0.0, abs=1e-8)


def test_windows(trajectory):
    def above_equator(s):
        return s.position()[2]

    windows = trajectory.find_windows(above_equator)
    assert len(windows) == 1



#  Copyright (c) 2024. Helge Eichhorn and the LOX contributors
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, you can obtain one at https://mozilla.org/MPL/2.0/.

import lox_space as lox
import numpy as np
import numpy.testing as npt
import pytest

ISS_TLE = """ISS (ZARYA)
1 25544U 98067A   24170.37528350  .00016566  00000+0  30244-3 0  9996
2 25544  51.6410 309.3890 0010444 339.5369 107.8830 15.49495945458731
"""


def test_sgp4():
    sgp4 = lox.SGP4(ISS_TLE)
    t1 = sgp4.time() + lox.TimeDelta.from_minutes(92.821)
    s1 = sgp4.propagate(t1)
    k1 = s1.to_keplerian()
    assert k1.orbital_period().to_decimal_seconds() == pytest.approx(
        92.821 * 60, rel=1e-4
    )


def test_ground(provider):
    lat = np.radians(40.4527)
    lon = np.radians(-4.3676)
    tai = lox.UTC.from_iso("2022-01-31T23:00:00").to_tai()
    loc = lox.GroundLocation(lox.Origin("Earth"), lon, lat, 0.0)
    ground = lox.GroundPropagator(loc, provider)
    expected = np.array([-1760.0326388434466, 4509.415893243589, 4106.374974141291])
    actual = ground.propagate(tai).position()
    npt.assert_allclose(actual, expected, rtol=1e-4)



#  Copyright (c) 2024. Helge Eichhorn and the LOX contributors
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, you can obtain one at https://mozilla.org/MPL/2.0/.

import itertools
import pytest

import lox_space as lox


@pytest.mark.benchmark()
def test_visibility_benchmark(provider, oneweb, estrack):
    mask = lox.ElevationMask.fixed(0)
    t0 = next(iter(oneweb.values())).states()[0].time()
    times = [t0 + t for t in lox.TimeDelta.range(0, 86400, 3000)]

    passes = {}

    for (gs_name, gs), (sc_name, sc) in filter(
        lambda pair: isinstance(pair[0][1], lox.GroundLocation)
        and isinstance(pair[1][1], lox.Trajectory),
        itertools.combinations(itertools.chain(estrack.items(), oneweb.items()), 2),
    ):
        assert isinstance(gs, lox.GroundLocation)
        assert isinstance(sc, lox.Trajectory)

        if not sc_name in passes:
            passes[sc_name] = {}

        passes[sc_name][gs_name] = lox.visibility(times, gs, mask, sc, provider)

    assert len(passes) == len(oneweb)
    for sc_passes in passes.values():
        assert len(sc_passes) == len(estrack)
